/*******************************************************************
 *
 *	File:		ArloApp.cpp
 *
 *	Author:		Peter van Sebille (peter@yipton.demon.co.uk)
 *
 *	(c) Copyright 2001, Peter van Sebille
 *	All Rights Reserved
 *
 *******************************************************************/

#include <e32std.h>
#include <e32cons.h>
#include <f32file.h>
#include <e32svr.h>
#include <w32std.h>

#include "RArlo.h"
#include "CArloController.h"


//#define DEBUG_PRINTS_PLEASE

#ifdef DEBUG_PRINTS_PLEASE
CConsoleBase*		sConsole = NULL;
#define DBGPRINT(x) sConsole->Printf x
#define DBGPRESSKEY(x) sConsole->Printf x;sConsole->Getch()
#else
#define DBGPRINT(x) /**/
#define DBGPRESSKEY(x) /**/
#endif


#ifdef __WINS__

#define EKERN_EXE_FILENAME		_L("c:\\test\\ekern.exe")
#define ECOMM_LDD_FILENAME		_L("c:\\test\\ecomm.ldd")
#define ARLO_LDD_FILENAME_ON_C	_L("c:\\test\\arloc.ldd")
#define ARLO_LDD_FILENAME_ON_D	_L("c:\\test\\arlod.ldd")

#else

#define EKERN_EXE_FILENAME		_L("z:\\system\\libs\\ekern.exe")
#define ECOMM_LDD_FILENAME		_L("z:\\system\\libs\\ecomm.ldd")
#define ARLO_LDD_FILENAME_ON_C	_L("c:\\system\\libs\\arlo.ldd")
#define ARLO_LDD_FILENAME_ON_D	_L("d:\\system\\libs\\arlo.ldd")

#endif


static TBool RunsAsSystemBootLoaderL(RWsSession& aWsSession)
{
	TInt			id=0;
	TInt			count = 0;

		/*
		 * We count the number of named window groups.
		 * If we run as a boot loader, the count is 1, otherwise there about 10 of them
		 */
	while (id >= 0)
	{
		id = aWsSession.FindWindowGroupIdentifier(id , _L("*"));
		if (id > 0)
		{
			TThreadId threadId;
			if (aWsSession.GetWindowGroupClientThreadId(id , threadId) == KErrNone)
			{
				RThread	thread;
				if (thread.Open(threadId) == KErrNone)
				{
					thread.Close();
					count++;
//					iConsole.Printf(_L("Threads: thread=%S, id=%d\n"), &name, id);
				}
			}
		}
	}
	return (count == 1);
}

static TBool RunsAsSystemBootLoader(RWsSession& aWsSession)
{
	TBool	runsAsSystemBootLoader = EFalse;
#ifndef __WINS__
	TRAPD(err, (runsAsSystemBootLoader = RunsAsSystemBootLoaderL(aWsSession)));
#endif
	return runsAsSystemBootLoader;
}


static TInt GetFileUids(const TDesC& aFileName, TUint& aUid0, TUint& aUid1, TUint& aUid2)
{
	RFs				fs;
	
	TInt err= fs.Connect();
	if (!err)
	{
		RFile		file;
		err = file.Open(fs, aFileName, 0);
		if (!err)
		{
			TBuf8<12>	uids;
			err = file.Read(uids);
			if (!err)
			{
				TUint*	p = (TUint*) uids.Ptr();
				aUid0 = p[0];
				aUid1 = p[1];
				aUid2 = p[2];
			}
			file.Close();
		}
		fs.Close();
	}
	return err;
}


/*
 * What we need to patch is:
 * 1) The first 2 UIDs in the arlo image file. For that we simply copy the 2 UIDs from
 * ecomm.ldd
 * 2) Arlo links against EKERN.EXE. Somewhere in the arlo image it says:
 * "EKERN[1000008b].EXEEUSER[100000c1].DLL". This is the section used by the loader
 * to determine against which .dlls this image has linked. The UID that is specified
 * between square brackets is the 3rd UID of that according binary image. In our case
 * the EKERN UIDs are unique per platform.
 */
TBool PatchDataIfNeeded(TDes8& aData, TUint aEkernUid, TUint aEcommUid0, TUint aEcommUid1)
{
	TBool	isChanged = EFalse;
	TUint*	uid = (TUint*) aData.Ptr();
	if (uid[0] != aEcommUid0)
	{
		uid[0] = aEcommUid0;
		isChanged = ETrue;
	}
	if (uid[1] != aEcommUid1)
	{
		uid[1] = aEcommUid1;
		isChanged = ETrue;
	}

	TInt pos = aData.Find(_L("EKERN["));
	if (pos != KErrNotFound)
	{
		pos += 6 + 7;
		TUint8*	data = (TUint8*) aData.Ptr() + pos;
		for (TInt i=0 ; i<8 ; i++)
		{
			TUint	c = (aEkernUid & 0xf);
			c += (c<10) ? '0' : ('a' - 10);
			if (*data != (TUint8) c)
			{
				*data = (TUint8) c;
				isChanged = ETrue;
			}
			data--;
			aEkernUid >>= 4;
		}
	}
	return isChanged;
}

static TBool PatchArloImageIfNeededL(const TDesC& aFilename, TUint aEkernUid, TUint aEcommUid0, TUint aEcommUid1)
{
	TBool		isChanged = EFalse;
	
	RFs		fsSession;
	User::LeaveIfError(fsSession.Connect());
	CleanupClosePushL(fsSession);

	RFile file;
	User::LeaveIfError(file.Open(fsSession, aFilename, EFileWrite));
	CleanupClosePushL(file);

	TInt fileSize;
	User::LeaveIfError(file.Size(fileSize));
	HBufC*	buf = HBufC::NewLC(fileSize);

	TBuf<32>	info;
	info.Format(_L("File size=%d"), fileSize);

	TPtr	ptr = buf->Des();
	User::LeaveIfError(file.Read(ptr));

	isChanged = PatchDataIfNeeded(ptr, aEkernUid, aEcommUid0, aEcommUid1);
	if (isChanged)
	{
			/*
			 * Unbelievable!
			 * file.Write(0, *buf) works fine on the Series5mx + the Series 7 but fails
			 * to work correctly on the Series 5. What gives !?
			 */
		User::LeaveIfError(file.Write(0, ptr));
	}

	CleanupStack::PopAndDestroy(3);	// buf, file, fsSession

	return isChanged;
}


static void PatchArloDriverIfNeededL()
{
	TUint	ekernUid, ecommUid0, ecommUid1, dummy;

	User::LeaveIfError(GetFileUids(EKERN_EXE_FILENAME, dummy, dummy, ekernUid));
	User::LeaveIfError(GetFileUids(ECOMM_LDD_FILENAME, ecommUid0, ecommUid1, dummy));

	TBool	isChanged = EFalse;
	TRAPD(err, isChanged = PatchArloImageIfNeededL(ARLO_LDD_FILENAME_ON_C, ekernUid, ecommUid0, ecommUid1));
	if (!err && isChanged)
	{
		User::InfoPrint(_L("Arlo driver successfully patched for this platform\n"));
	}

	TRAPD(err1, isChanged = PatchArloImageIfNeededL(ARLO_LDD_FILENAME_ON_D, ekernUid, ecommUid0, ecommUid1));
	if (!err1 && isChanged)
	{
		User::InfoPrint(_L("Arlo driver successfully patched for this platform\n"));
	}
}

static void PatchArloDriverIfNeeded()
{
	TRAPD(err, PatchArloDriverIfNeededL());
}

static void DumpFileUids(CConsoleBase* aConsole, const TDesC& aName, const TDesC& aFileName)
{
	TUint	uid0, uid1, uid2;

	TInt err = GetFileUids(aFileName, uid0, uid1, uid2);
	if (!err)
		aConsole->Printf(_L("%S: uid[0]=0x%x, uid[1]=0x%x, uid[2]=0x%x\n"), &aName, uid0, uid1, uid2);
}


static void DumpUidInfo(CConsoleBase* aConsole)
{
	DumpFileUids(aConsole, _L("Ekern"), EKERN_EXE_FILENAME);
	DumpFileUids(aConsole, _L("EComm"), ECOMM_LDD_FILENAME);
}


static TInt LoadArloDriver(CConsoleBase* aConsole, RArLo& aArlo)
{
	aConsole->Printf(_L("Loading Arlo device driver\n"));

	TInt err = User::LoadLogicalDevice(KArloDriverName);
	if ((err!=KErrNone) && (err!=KErrAlreadyExists))
	{
		aConsole->Printf(_L("Error %d while loading logical device\n"), err);
		DumpUidInfo(aConsole);
	}
	else
	{
		if (err == KErrAlreadyExists)
		{
			aConsole->Printf(_L("  (driver was already loaded)\n"));
		}
		aConsole->Printf(_L("Opening the channel\n"));
		err = aArlo.Open(aArlo.VersionRequired());
		if (err) 
		{
			aConsole->Printf(_L("Error %d while opening channel to Arlo device driver\n"), err);
			User::FreeLogicalDevice(KArloDriverName);
		}
	}
	DumpUidInfo(aConsole);

	return err;
}


void TestVideo()
{
}




static void ArloMain1L(CConsoleBase* aConsole)
{
	RArLo			arloDevice;

	PatchArloDriverIfNeeded();

	RFs		fsSession;
	User::LeaveIfError(fsSession.Connect());
	CleanupClosePushL(fsSession);

	RWsSession		aWsSession;
#ifndef __WINS__
	User::LeaveIfError(aWsSession.Connect());
	CleanupClosePushL(aWsSession);
#endif

	User::LeaveIfError(LoadArloDriver(aConsole, arloDevice));

	CArloController*	arloController = CArloController::NewLC(*aConsole, arloDevice, fsSession, aWsSession);

	arloController->DoTheArloThingL(RunsAsSystemBootLoader(aWsSession));

	CleanupStack::PopAndDestroy();		// arloController

	arloDevice.Close();
	User::FreeLogicalDevice(KArloDriverName);
	
#ifndef __WINS__
	CleanupStack::PopAndDestroy();	// aWsSession;
#endif
	CleanupStack::PopAndDestroy();	// fsSession;
}


void ArloMainL()
{
	CConsoleBase*	console = Console::NewL(_L("ARLO"), TSize(KDefaultConsWidth,KDefaultConsHeight));
	CleanupStack::PushL(console);

#ifdef DEBUG_PRINTS_PLEASE
	sConsole = console;
#endif

	// start by creating a console, so that we can display any errors
	TRAPD(err, ArloMain1L(console));
	if (err)
	{
		console->Printf(_L("Oops, something failed: leave code=%d\n"), err);
		console->Printf(_L("Press a key\n"));
		console->Getch();
	}

	CleanupStack::PopAndDestroy();		// console
}


GLDEF_C TInt E32Main() // main function called by E32
{
	__UHEAP_MARK;

	CTrapCleanup* cleanup=CTrapCleanup::New();

	TRAPD(error, ArloMainL());                     

	__ASSERT_ALWAYS(!error, User::Panic(_L("LINUX"), error));

	delete cleanup; // destroy clean-up stack

	__UHEAP_MARKEND;

	return 0; // and return
}

