/*******************************************************************
 *
 *	File:		CConsoleEditor.cpp
 *
 *	Author:		Peter van Sebille (peter@yipton.demon.co.uk)
 *
 *	(c) Copyright 2001, Peter van Sebille
 *	All Rights Reserved
 *
 *******************************************************************/

#include "CConsoleEditor.h"

CConsoleEditor* CConsoleEditor::NewL(CConsoleBase* aConsole)
{
	CConsoleEditor*	self = new(ELeave)CConsoleEditor(aConsole);
	CleanupStack::PushL(self);
	self->ConstructL();
	CleanupStack::Pop();		// self
	return self;
}


void CConsoleEditor::ConstructL()
{
	iScreenSize = iConsole->ScreenSize();
	iOverWrite = EFalse;
}




void CConsoleEditor::EditLine(const TDesC& aPrompt, TDes& aString)
{
	iConsole->Printf(aPrompt);
	iHome = iConsole->CursorPos();
	iConsole->SetCursorHeight(iOverWrite ? 100 : 10);

	iMaxText = iScreenSize.iWidth - iHome.iX - 3;
	aString.Zero();

	TBool finished = EFalse;
	while (!finished)
	{
		TChar	c = iConsole->Getch();
		switch (c)
		{
			case EKeyHome:
			{
				iConsole->SetPos(iHome.iX);
				break;
			}

			case EKeyEnd:
			{
				iConsole->SetPos(iHome.iX + aString.Length());
				break;
			}

			case EKeyDelete:
			{
				TInt pos = iConsole->WhereX() - iHome.iX;
				if (pos<aString.Length())
				{
					aString.Delete(pos, 1);
					TInt	newPos = iConsole->WhereX();
					iConsole->Write(aString.Mid(pos));
					iConsole->Write(_L(" "));
					iConsole->SetPos(newPos);
				}
				break;
			}

			case EKeyInsert:
				iOverWrite = iOverWrite ? (TBool) EFalse : (TBool) ETrue;
				iConsole->SetCursorHeight(iOverWrite ? 100 : 10);
				break;

			case EKeyEnter:
				iConsole->SetCursorHeight(0);
				iConsole->Write(_L("\n"));
				finished = ETrue;
				break;

			case EKeyBackspace:
			{
				TInt pos = iConsole->WhereX() - iHome.iX;
				if (pos > 0)
				{
					aString.Delete(pos - 1, 1);
					TInt	newPos = iConsole->WhereX()-1;
					iConsole->SetPos(newPos);
					iConsole->Write(aString.Mid(pos-1));
					iConsole->Write(_L(" "));
					iConsole->SetPos(newPos);
				}
				break;
			}
			
			case EKeyLeftArrow:
			{
				TInt pos = iConsole->WhereX() - iHome.iX;
				if (pos > 0)
				{
					TInt	newPos = iConsole->WhereX()-1;
					iConsole->SetPos(newPos);
				}
				break;
			}

			case EKeyRightArrow:
			{
				TInt pos = iConsole->WhereX() - iHome.iX;
				if (pos < aString.Length())
				{
					TInt	newPos = iConsole->WhereX()+1;
					iConsole->SetPos(newPos);
				}
				break;
			}

			default:
			{
				if (!c.IsPrint())
					break;

				TBuf<1>		cString;
				cString.Append((TUint8) c);

				TInt pos = iConsole->WhereX() - iHome.iX;
				if (pos == aString.Length())
				{
					if (aString.Length() < iMaxText)
					{
						aString.Append(c);
						iConsole->Printf(cString);
					}
				}
				else if (iOverWrite)
				{
					aString.Replace(pos, 1, cString);
					iConsole->Write(cString);
				}
				else
				{
					if (aString.Length() < iMaxText)
					{
						TInt	newPos = iConsole->WhereX()+1;
						aString.Insert(pos, cString);
						iConsole->Write(aString.Mid(pos));
						iConsole->SetPos(newPos);
					}
				}
			}
		}
//		RDebug::Print(_L("text=%S"), &aString);
	}
}
