/*******************************************************************
 *
 *	File:		Arlo.cpp
 *
 *	Author:		Peter van Sebille (peter@yipton.demon.co.uk)
 *				Based on code from Otfried Cheong
 *
 *	(c) Copyright 2001, Peter van Sebille
 *	All Rights Reserved
 *
 *******************************************************************/

#include <e32std.h>
#include <e32base.h>
#include <e32hal.h>
#include <e32rom.h>
#include <e32ver.h>
#include <e32svr.h>

#include "ekern.h"
#include "arlo.h"

class TCapsArLoV1
{
public:
	TVersion version;
};


// version information
const TInt KMajorVersionNumber=0;
const TInt KMinorVersionNumber=0;
const TInt KBuildVersionNumber=0;


class DArLoDriver : public DLogicalDevice
{
public:
	DArLoDriver();
	virtual TInt Install();
	virtual void GetCaps(TDes8 &aDes) const;
	virtual DLogicalChannel *CreateL();

		// for some reason the default Remove returns KErrNotSupported
		// if we return KErrNon in Remove the machine reboots :-(
//	virtual TInt Remove(void){return KErrGeneral;}
};


class DArLo : public DLogicalChannel
{
public:
	DArLo(DLogicalDevice *aDevice);
protected:
	virtual void DoCancel(TInt aReqNo);
	virtual void DoRequest(TInt aReqNo,TAny *a1,TAny *a2);
	virtual TInt DoControl(TInt aFunction,TAny *a1,TAny *a2);
	TInt OriginalBoot(TInt aFunction, TAny* a1, TAny* a2);
	TInt BootS7(TAny* a1, TAny* a2);
};



DArLo::DArLo(DLogicalDevice *aDevice)
  : DLogicalChannel(aDevice)
{
//  SetBehaviour(0); // indicate we do not support asynchronous requests
}

void DArLo::DoCancel(TInt /* aReqNo */)
{
	INT3;
}

void DArLo::DoRequest(TInt /*aReqNo*/,TAny* /*a1*/,TAny* /*a2*/)
{
	INT3;
}




void FiddleDomainAccess()
{
#ifdef __MARM__
	asm("mov r0, #0xffffffff    ");
	asm("mcr p15, 0, r0, c3, c0 ");
#endif
}

#define __NAKED__ __declspec( naked )

#ifdef __MARM__
__NAKED__ TUint GetProcessorId()
{
	asm("  mrc	p15, 0, r0, c0, c0");		// @ get Processor ID
//	asm("mrs r0, cpsr");
	asm("  mov  pc, lr");  // return
}
#else
TUint GetProcessorId()
{
	return 0xf;
}
#endif


__NAKED__ TInt DArLo::OriginalBoot(TInt aFunction, TAny* a1, TAny* a2)
{
#ifdef __MARM__
  // handle any synchronous request function
  // can trash r0-r3 and ip
  // on entry:
  //   r0=this
  //   r1=aFunction
  //   r2=a1 (PhysMapAdr)
  //   r3=a2 (BootAdr)
  // on exit:
  //   r0=result
  //                                  

  // turn off interrupts
  asm("  mrs r0, cpsr");
  asm("  bic r0, r0, #0xdf");
  asm("  orr r0, r0, #0xd3");  // disable FIQ, IRQ, set SVC mode
  asm("  msr cpsr,r0");

  // load r0 with physical map address
  asm("  mov r0, r2");

  // turn off MMU (write zero into MMU register)
  asm("  mov R1,#070");      // not quite zero---don't want 26 bit mode
  asm("  mcr p15,0,R1,c1,c1");

  // we have two instructions here before hell breaks lose!
  // jump to R3
  asm("  mov pc,R3");
  // will never get here
#endif
}

__NAKED__ TInt DArLo::BootS7(TAny* a1, TAny* a2)
{
#ifndef __WINS__
// on entry:
//		r0=this
//		r1=a1 (PhysMapAdr)
//		r2=a2 (BootAdr)

	// bootstrap code wants the PhysMapAdr in r0
	asm("mov r0, r1");

		/*
		 * set watchdeadline as far as possible in the future
		 */
//	TUint*	p = (TUint*) 0x58008000;
//	TUint	current = p[4];					// get current tick
//	p[3] = current-1;						// set deadline

	asm("mov r3, #0x58000000");
	asm("add r3, r3, #0x8000");
	asm("ldr r4, [r3, #16]");
	asm("sub r4, r4, #1");
	asm("str r4, [r3, #12]");

	asm("mrs r3, cpsr");
	asm("bic r3, r3, #0x1f");				// clear mode
	asm("orr r3, r3, #0x000000d3");			// disable IRQs and FIQs, SVC mode
	asm("msr cpsr, r3");

	asm("mov r3, #0");
	asm("mcr p15, 0, r3, c7, c7, 0");		// invalidate I,D caches
	asm("mcr p15, 0, r3, c7, c10, 4");		// drain WB
	asm("mcr p15, 0, r3, c8, c7, 0");		// invalidate I & D TLBs

	asm("mrc p15, 0, r3, c1, c0, 0");		// read ctrl register
	asm("bic r3, r3, #0x0000f");			// ............wcam
	asm("bic r3, r3, #0x00100");			// .......s........
	asm("bic r3, r3, #0x01000");			// ...i............

	asm("mcr p15, 0, r3, c1, c0, 0");		// write ctrl register

	asm("mov r3, r3");						// one
	asm("mov r3, r3");						// two

	asm("mov pc, r2");						// three, bye bye
	// will never get here
#endif
}


TInt DArLo::DoControl(TInt aFunction, TAny* a1, TAny* a2)
{
	INT3;

	switch (aFunction)
	{
		case KArloLddBootS5:
		case KArloLddBootS5mx:
			OriginalBoot(aFunction, a1, a2);
			break;

		case KArloLddBootS7:
			FiddleDomainAccess();
			BootS7(a1, a2);
			break;

		case KArloLddExamineArm4:
			FiddleDomainAccess();			// fall through
		case KArloLddExamineS5:
		{
			TUint*	val = (TUint*) a1;
			(*val) = *((TUint32*) *val);
			break;
		}

		case KArloLddProcId:
		{
			TUint*		p = (TUint*) a1;
			*p = GetProcessorId();
			break;
		}

		default:
			break;

	}
	return KErrNone;
}




DArLoDriver::DArLoDriver()
{
}

TInt DArLoDriver::Install()
{
	INT3;
	TPtrC name=KArloDriverName;       // our name, used by User::FreeLogicalDevice
	return SetName(&name);       // set our name and return error code
}

void DArLoDriver::GetCaps(TDes8 &aDes) const
{
	TVersion  version(KMajorVersionNumber, KMinorVersionNumber, KBuildVersionNumber);

	INT3;
	// standard capability-descriptor pattern
	TCapsArLoV1 b; // a structure
	b.version=version; // fill it with information
	aDes.FillZ(aDes.MaxLength()); // zero-pad the caller's descriptor
	aDes.Copy((TUint8 *)&b,Min(aDes.MaxLength(),sizeof(b)));
}

DLogicalChannel* DArLoDriver::CreateL()
{
	INT3;
	return new(ELeave) DArLo(this);
}


EXPORT_C DLogicalDevice *CreateLogicalDevice()
{
	INT3;
	DLogicalDevice *arlo = new(ELeave) DArLoDriver;
	return arlo;
}


// DLL entry point, required by all DLLs,
// called on thread or process attach/detach
GLDEF_C TInt E32Dll(TDllReason /*aReason*/)
{
  return KErrNone; // do nothing, return no error
}

